/*
 * Display control for Huge RGB Display
 */

#include <stdint.h>
#include <stdbool.h>

#include "tm4c129encpdt.h"

#include "inc/hw_gpio.h"
#include "inc/hw_memmap.h"
#include "inc/hw_sysctl.h"
#include "inc/hw_types.h"
#include "inc/hw_ints.h"
#include "inc/hw_udma.h"
#include "inc/hw_ssi.h"

#include "driverlib/debug.h"
#include "driverlib/adc.h"
#include "driverlib/udma.h"
#include "driverlib/fpu.h"
#include "driverlib/interrupt.h"
#include "driverlib/gpio.h"
#include "driverlib/pin_map.h"
#include "driverlib/uart.h"
#include "driverlib/rom.h"
#include "driverlib/ssi.h"
#include "driverlib/sysctl.h"
#include "driverlib/systick.h"
#include "driverlib/timer.h"
#include "driverlib/hibernate.h"
#include "driverlib/pwm.h"

#include "hal.h"
#include "dispctl.h"

// OE scale factors for calculated pulse widths
// Red/green/blue colour balance altered here; global brightness setting affects whole panel
// Subfield and whole-frame ABL is calculated as an additional factor
uint32_t oe_ints_global = 1024;		// 1024 = max intensity
uint32_t oe_ints_r	 		= 950;		// ^
uint32_t oe_ints_g		 	= 500;		// ^
uint32_t oe_ints_b		 	= 1024;		// ^

// OE pulse widths - as driven to panel
uint32_t oe_width_r, oe_width_g, oe_width_b, oe_width_all, oe_base;

// OE states
uint32_t oe_r_state, oe_g_state, oe_b_state;

// Generate OE?  If cleared, no OE pulses are generated for this subfield,
// which reduces quiescent power
uint32_t oe_gen;

// Pulse widths of each subfield as a base figure
const uint16_t subfield_widths[DISP_SUBFIELDS] = {
		(DISPCTL_OE_MAX_PULSE / 512),											// 3
		(DISPCTL_OE_MAX_PULSE / 256),											// 4
		(DISPCTL_OE_MAX_PULSE / 128),											// 5
		(DISPCTL_OE_MAX_PULSE / 64),											// 6
		(DISPCTL_OE_MAX_PULSE / 32),											// 7
		(DISPCTL_OE_MAX_PULSE / 16),											// 8
		(DISPCTL_OE_MAX_PULSE / 8),												// 9
		(DISPCTL_OE_MAX_PULSE / 4),												// 10
		(DISPCTL_OE_MAX_PULSE / 2),												// 11
		 DISPCTL_OE_MAX_PULSE,														// 12
};

// Dividers for said subfields (must be a power of two)
const uint16_t subfield_divider[DISP_SUBFIELDS] = {
	//2048, 1024, 512, 256, 128, 64, 32, 16, 8, 4
	//4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048
	64, 128,
	256, 512, 1024, 2048, 
	4096, 8192, 16384, 32768
};

// Current subfield from 0 - (DISP_SUBFIELDS-1)
uint32_t cur_subfield = 0;

// Frame counter (32 bit, overflowing)
uint32_t disp_frame = 0;

// LED divider for frame sync output.
uint32_t disp_led_count = 0;
uint32_t disp_led_state = 0;

// X, Y, height and latch lookup for modules
struct disp_xyoff_t disp_xyoff[DISP_NUM_SUPERBLOCKS] = {
	/*
	{ 	0, 	0, 	8,	 4  },
	{  16, 	0, 	8,	 18 },
	{  32, 	0, 	8,	 0  },
	{  48, 	0, 	8,	 9  },
	{  64, 	0, 	8,	 8  },
	
	{ 	0, 16,  7,	 17 }, 
	{  16, 16,  7,	 3  }, 
	{  32, 16,  7,	 5  }, 
	{  48, 16,  7,	 7  }, 
	{  64, 16,  7,	 6  },
	
	{ 	0, 32, 	8,	10  },
	{  16, 32, 	8,	12  },
	{  32, 32, 	8,	14  }, 
	{  48, 32, 	8,	16  },
	{  64, 32, 	8,	1   },
	
	{ 	0, 48,  7,	11  }, 
	{  16, 48,  7,	13  }, 
	{  32, 48,  7,	15  },
	{  48, 48,  7,	2   }, 
	{  64, 48,  7,	19  }
	*/
	
	{ 	0, 	0, 	8,	 4  }, // correct
	{  16, 	0, 	8,	 16 },
	{  32, 	0, 	8,	 0  }, // zero correct
	{  48, 	0, 	8,	 14 },
	{  64, 	0, 	8,	 11 },
	
	{ 	0, 16,  7,	 17 }, 
	{  16, 16,  7,	 13 }, 
	{  32, 16,  7,	 19 }, 
	{  48, 16,  7,	 18 }, 
	{  64, 16,  7,	 6  },
	
	{ 	0, 30, 	8,	 12 },
	{  16, 30, 	8,	 10 },
	{  32, 30, 	8,	 9  }, 
	{  48, 30, 	8,	 7  },
	{  64, 30, 	8,	 1  },
	
	{ 	0, 46,  7,	 8  }, 
	{  16, 46,  7,	 5  }, 
	{  32, 46,  7,	 15 },
	{  48, 46,  7,	 2  }, 
	{  64, 46,  7,	 3  }
};

// Latch port lookup
const struct disp_latch_lut_t disp_latch_lut[DISP_NUM_LATCHES] = {
	// PORT BASE,  			PIN MASK,   	PIN INVERSE MASK
	{ GPIO_PORTN_BASE,	GPIO_PIN_3,	  ~GPIO_PIN_3	},			// Latch 0
	{ GPIO_PORTM_BASE,	GPIO_PIN_4,	  ~GPIO_PIN_4	},			// Latch 1
	{ GPIO_PORTN_BASE,	GPIO_PIN_2,	  ~GPIO_PIN_2	},			// Latch 2
	{ GPIO_PORTM_BASE,	GPIO_PIN_5,	  ~GPIO_PIN_5	},			// Latch 3
	{ GPIO_PORTN_BASE,	GPIO_PIN_1,	  ~GPIO_PIN_1	},			// Latch 4
	{ GPIO_PORTM_BASE,	GPIO_PIN_6,	  ~GPIO_PIN_6	},			// Latch 5
	{ GPIO_PORTN_BASE,	GPIO_PIN_0,	  ~GPIO_PIN_0	},			// Latch 6
	{ GPIO_PORTM_BASE,	GPIO_PIN_7,	  ~GPIO_PIN_7	},			// Latch 7
	{ GPIO_PORTN_BASE,	GPIO_PIN_5,	  ~GPIO_PIN_5	},			// Latch 8
	{ GPIO_PORTD_BASE,	GPIO_PIN_1,	  ~GPIO_PIN_1	},			// Latch 9
	{ GPIO_PORTN_BASE,	GPIO_PIN_4,	  ~GPIO_PIN_4	},			// Latch 10
	{ GPIO_PORTL_BASE,	GPIO_PIN_3,	  ~GPIO_PIN_3	},			// Latch 11
	{ GPIO_PORTL_BASE,	GPIO_PIN_5,	  ~GPIO_PIN_5	},			// Latch 12
	{ GPIO_PORTP_BASE,	GPIO_PIN_5,	  ~GPIO_PIN_5	},			// Latch 13
	{ GPIO_PORTK_BASE,	GPIO_PIN_4,	  ~GPIO_PIN_4	},			// Latch 14
  { GPIO_PORTP_BASE,	GPIO_PIN_4,	  ~GPIO_PIN_4	},			// Latch 15
	{ GPIO_PORTH_BASE,	GPIO_PIN_3,	  ~GPIO_PIN_3	},			// Latch 16
	{ GPIO_PORTH_BASE,	GPIO_PIN_2,	  ~GPIO_PIN_2	},			// Latch 17
	{ GPIO_PORTH_BASE,	GPIO_PIN_1,	  ~GPIO_PIN_1	},			// Latch 18
	{ GPIO_PORTH_BASE,	GPIO_PIN_0,	  ~GPIO_PIN_0	},			// Latch 19
};

// Address bitmask lookup
const struct disp_addr_lut_t disp_addr_lut[DISP_NUM_INNER_MODULES] = {
	// Index 0
	{
		DISPCTL_ADDR_BIT0_PORT,
		~DISPCTL_ADDR_BIT0_MASK,		// 0
		DISPCTL_ADDR_BIT1_PORT,
		~DISPCTL_ADDR_BIT1_MASK,		// 0
		DISPCTL_ADDR_BIT2_PORT,
		~DISPCTL_ADDR_BIT2_MASK,		// 0
		DISPCTL_ADDR_BIT0_MASK,			// writemask
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK,			// writemask
	},
	
	// Index 1
	{
		DISPCTL_ADDR_BIT0_PORT,
		DISPCTL_ADDR_BIT0_MASK,			// 1
		DISPCTL_ADDR_BIT1_PORT,
		~DISPCTL_ADDR_BIT1_MASK,		// 0
		DISPCTL_ADDR_BIT2_PORT,
		~DISPCTL_ADDR_BIT2_MASK,		// 0
		DISPCTL_ADDR_BIT0_MASK,			// writemask
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK			// writemask
	},
		
	// Index 2
	{
		DISPCTL_ADDR_BIT0_PORT,
		~DISPCTL_ADDR_BIT0_MASK,		// 0
		DISPCTL_ADDR_BIT1_PORT,
		DISPCTL_ADDR_BIT1_MASK,			// 1
		DISPCTL_ADDR_BIT2_PORT,
		~DISPCTL_ADDR_BIT2_MASK,		// 0
		DISPCTL_ADDR_BIT0_MASK,			// writemask 
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK			// writemask
	},
	
	// Index 3
	{
		DISPCTL_ADDR_BIT0_PORT,
		DISPCTL_ADDR_BIT0_MASK,			// 1
		DISPCTL_ADDR_BIT1_PORT,
		DISPCTL_ADDR_BIT1_MASK,			// 1
		DISPCTL_ADDR_BIT2_PORT,
		~DISPCTL_ADDR_BIT2_MASK,		// 0
		DISPCTL_ADDR_BIT0_MASK,			// writemask
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK			// writemask
	},
		
	// Index 4
	{
		DISPCTL_ADDR_BIT0_PORT,
		~DISPCTL_ADDR_BIT0_MASK,		// 0
		DISPCTL_ADDR_BIT1_PORT,
		~DISPCTL_ADDR_BIT1_MASK,		// 0
		DISPCTL_ADDR_BIT2_PORT,
		DISPCTL_ADDR_BIT2_MASK,			// 1
		DISPCTL_ADDR_BIT0_MASK,			// writemask
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK			// writemask
	},
		
	// Index 5
	{
		DISPCTL_ADDR_BIT0_PORT,
		DISPCTL_ADDR_BIT0_MASK,			// 1
		DISPCTL_ADDR_BIT1_PORT,
		~DISPCTL_ADDR_BIT1_MASK,		// 0
		DISPCTL_ADDR_BIT2_PORT,
		DISPCTL_ADDR_BIT2_MASK,			// 1
		DISPCTL_ADDR_BIT0_MASK,			// writemask
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK			// writemask
	},
		
	// Index 6
	{
		DISPCTL_ADDR_BIT0_PORT,
		~DISPCTL_ADDR_BIT0_MASK,		// 0
		DISPCTL_ADDR_BIT1_PORT,
		DISPCTL_ADDR_BIT1_MASK,			// 1
		DISPCTL_ADDR_BIT2_PORT,
		DISPCTL_ADDR_BIT2_MASK,			// 1
		DISPCTL_ADDR_BIT0_MASK,			// writemask
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK			// writemask
	},
		
	// Index 7
	{
		DISPCTL_ADDR_BIT0_PORT,
		DISPCTL_ADDR_BIT0_MASK,			// 1
		DISPCTL_ADDR_BIT1_PORT,
		DISPCTL_ADDR_BIT1_MASK,			// 1
		DISPCTL_ADDR_BIT2_PORT,
		DISPCTL_ADDR_BIT2_MASK,			// 1
		DISPCTL_ADDR_BIT0_MASK,			// writemask
		DISPCTL_ADDR_BIT1_MASK,			// writemask
		DISPCTL_ADDR_BIT2_MASK			// writemask
	}
};

// Current display refresh rate in Hz.  Currently fixed; future feature should allow for dynamic refresh
// rates.
uint32_t disp_refresh_rate = DISP_REFRESH_RATE_DEFAULT;

// Display buffers
struct pix_rgbg_t disp_1[DISP_PIX_WIDTH * DISP_PIX_HEIGHT];
struct pix_rgbg_t disp_2[DISP_PIX_WIDTH * DISP_PIX_HEIGHT];
// struct pix_rgbg_t disp_out[DISP_PIX_WIDTH * DISP_PIX_HEIGHT];

// Current display pointer and working (scratchpad) display pointer,
// these are flipped after the display is flushed
struct pix_rgbg_t *disp_current_ptr;
struct pix_rgbg_t *disp_working_ptr;

// Current calculated subfield in byte packets
// Data is ordered in each byte:  Rx/G1/Bx/G2 Rx/G1/Bx/G2  - with each Rx/G1/G2/Bx being one subpixel 
uint8_t disp_cur_subfield[DISP_SUBFIELD_SIZE_BYTES];

// Gamma lookup table, map of 0-255 to 0-65535
const uint16_t def_gamma_lookup_12b[256] = { 
	0x0000,  0x0001,  0x0002,  0x0004,  0x0007,  0x000c,  0x0011,  0x0018,  
	0x0021,  0x002a,  0x0035,  0x0041,  0x004f,  0x005e,  0x006f,  0x0081,  
	0x0095,  0x00aa,  0x00c0,  0x00d9,  0x00f3,  0x010e,  0x012b,  0x014a,  
	0x016a,  0x018c,  0x01b0,  0x01d5,  0x01fc,  0x0225,  0x024f,  0x027c,  
	0x02aa,  0x02d9,  0x030b,  0x033e,  0x0373,  0x03aa,  0x03e3,  0x041d,  
	0x045a,  0x0498,  0x04d8,  0x051a,  0x055d,  0x05a3,  0x05ea,  0x0634,  
	0x067f,	 0x06cc,  0x071b,  0x076c,  0x07bf,  0x0814,  0x086b,  0x08c3,  
	0x091e,  0x097b,  0x09da,  0x0a3a,  0x0a9d,  0x0b01,  0x0b68,  0x0bd1,  
	0x0c3b,  0x0ca8,  0x0d16,  0x0d87,  0x0dfa,  0x0e6f,  0x0ee5,  0x0f5e,  
	0x0fd9,  0x1056,  0x10d5,  0x1157,  0x11da,  0x125f,  0x12e6,  0x1370,  
	0x13fc,  0x1489,  0x1519,  0x15ab,  0x163f,  0x16d5,  0x176e,  0x1808,  
	0x18a5,  0x1944,  0x19e5,  0x1a88,  0x1b2d,  0x1bd5,  0x1c7e,  0x1d2a,  
	0x1dd8,  0x1e88,  0x1f3a,  0x1fef,  0x20a6,  0x215f,  0x221a,  0x22d7,  
	0x2397,  0x2459,  0x251d,  0x25e3,  0x26ac,  0x2776,  0x2843,  0x2913,  
	0x29e4,  0x2ab8,  0x2b8e,  0x2c66,  0x2d41,  0x2e1e,  0x2efd,  0x2fde,  
	0x30c2,  0x31a8,  0x3290,  0x337b,  0x3468,  0x3557,  0x3648,  0x373c,  
	0x3832,  0x392b,  0x3a25,  0x3b22,  0x3c22,  0x3d24,  0x3e28,  0x3f2e,  
	0x4037,  0x4142,  0x424f,  0x435f,  0x4471,  0x4586,  0x469d,  0x47b6,  
	0x48d2,  0x49f0,  0x4b10,  0x4c33,  0x4d58,  0x4e7f,  0x4fa9,  0x50d5,  
	0x5204,  0x5335,  0x5469,  0x559f,  0x56d7,  0x5812,  0x594f,  0x5a8e,  
	0x5bd0,  0x5d15,  0x5e5b,  0x5fa5,  0x60f0,  0x623f,  0x638f,  0x64e2,  
	0x6638,  0x678f,  0x68ea,  0x6a47,  0x6ba6,  0x6d08,  0x6e6c,  0x6fd2,  
	0x713b,  0x72a7,  0x7415,  0x7586,  0x76f9,  0x786e,  0x79e6,  0x7b61,  
	0x7cdd,  0x7e5d,  0x7fdf,  0x8163,  0x82ea,  0x8474,  0x8600,  0x878e,  
	0x891f,  0x8ab2,  0x8c48,  0x8de1,  0x8f7c,  0x911a,  0x92ba,  0x945c,  
	0x9601,  0x97a9,  0x9953,  0x9b00,  0x9caf,  0x9e61,  0xa016,  0xa1cd,  
	0xa386,  0xa542,  0xa701,  0xa8c2,  0xaa86,  0xac4c,  0xae15,  0xafe0,  
	0xb1ae,  0xb37f,  0xb552,  0xb728,  0xb900,  0xbadb,  0xbcb8,  0xbe98,  
	0xc07b,  0xc260,  0xc448,  0xc632,  0xc81f,  0xca0f,  0xcc01,  0xcdf6,  
	0xcfee,  0xd1e8,  0xd3e4,  0xd5e4,  0xd7e6,  0xd9ea,  0xdbf1,  0xddfb,  
	0xe008,  0xe217,  0xe428,  0xe63d,  0xe853,  0xea6d,  0xec89,  0xeea8,  
	0xf0ca,  0xf2ee,  0xf515,  0xf73e,  0xf96a,  0xfb99,  0xfdca,  0xffef
};

// Uploaded gamma tables
uint16_t gamma_lookup_red[256];
uint16_t gamma_lookup_green[256];
uint16_t gamma_lookup_blue[256];

uint16_t *gamma_table_red;
uint16_t *gamma_table_green;
uint16_t *gamma_table_blue;

// Frame wait and sync, set by runtime software, cleared by interrupts.
// Must be volatile to avoid compiler optimisation.
volatile uint32_t disp_frame_wait = 0;
volatile uint32_t disp_frame_sync = 0;
volatile uint32_t disp_ptr_order = 0;

// ABL peak and average limits
uint32_t abl_peak_power = 24000;			// 24 watts peak
uint32_t abl_avg_power = 12000;				// 12 watts average
uint32_t calc_cur_power = 0;
uint32_t calc_avg_power = 0;
uint32_t abl_limit_oe = 0;

/*
 * Timer0AIntHandler:  OE timeout for Red OE
 */
void Timer0AIntHandler()
{
	TimerIntClear(TIMER0_BASE, TIMER_TIMA_TIMEOUT);
	
	if(oe_r_state == OE_STATE_GOLOW) {
		oe_r_state = OE_STATE_GOHIGH;
		
		// load timer
		TimerLoadSet(TIMER0_BASE, TIMER_A, oe_width_r);
		TimerEnable(TIMER0_BASE, TIMER_A);
		
		if(oe_gen) {
			DISPCTL_R_OE_ON();
		}
	} else if(oe_r_state == OE_STATE_GOHIGH) { 
		DISPCTL_R_OE_OFF();
	}
}
	
/*
 * Timer0BIntHandler:  OE timeout for Green OE
 */
void Timer0BIntHandler()
{
	TimerIntClear(TIMER0_BASE, TIMER_TIMB_TIMEOUT);
	
	if(oe_g_state == OE_STATE_GOLOW) {
		oe_g_state = OE_STATE_GOHIGH;
		
		// load timer
		TimerLoadSet(TIMER0_BASE, TIMER_B, oe_width_g);
		TimerEnable(TIMER0_BASE, TIMER_B);
		
		if(oe_gen) {
			DISPCTL_G_OE_ON();
		}
	} else if(oe_g_state == OE_STATE_GOHIGH) { 
		DISPCTL_G_OE_OFF();
	}
}
	
/*
 * Timer1AIntHandler:  OE timeout for Blue OE
 */
void Timer1AIntHandler()
{
	TimerIntClear(TIMER1_BASE, TIMER_TIMA_TIMEOUT);
	
	if(oe_b_state == OE_STATE_GOLOW) {
		oe_b_state = OE_STATE_GOHIGH;
		
		// load timer
		TimerLoadSet(TIMER1_BASE, TIMER_A, oe_width_b);
		TimerEnable(TIMER1_BASE, TIMER_A);
		
		if(oe_gen) {
			DISPCTL_B_OE_ON();
		}
	} else if(oe_b_state == OE_STATE_GOHIGH) { 
		DISPCTL_B_OE_OFF();
	}
}

/*
 * Timer1BIntHandler:  Overall OE timeout for subfield generation.
 */
void Timer1BIntHandler()
{
	TimerIntClear(TIMER1_BASE, TIMER_TIMB_TIMEOUT);
	disp_drive_subfield();
}

/*
 * Timer2AIntHandler:  Frame timer.
 */
void Timer2AIntHandler()
{
	uint32_t fr_power = 0, n;
	uint32_t abl_limit_peak = 1023, abl_limit_avg = 1023; 
	struct pix_rgbg_t *temp;
	
	TimerIntClear(TIMER2_BASE, TIMER_TIMA_TIMEOUT);
	
	// Sync frame?
	if(disp_frame_sync) {
		disp_ptr_order ^= 1;
		temp = disp_current_ptr;
		disp_current_ptr = disp_working_ptr;
		disp_working_ptr = temp;
		disp_frame_sync = 0;
	}
	
	// Run the ABL engine for this frame.  This works out the peak power of the frame,
	// based on the sum of pixels, and the average power over time.
	//
	// It will reduce the global brightness if either peak or average limits are exceeded.
	//
	// A separate ABL limit applies to individual subfields based on peak current within
	// a given subfield.
	/*
	for(n = 0; n < DISP_PIXELS; n++) {
		// TODO: account for relative brightness of R/G/B?
		fr_power += disp_current_ptr[n].r + disp_current_ptr[n].g1 + disp_current_ptr[n].b + disp_current_ptr[n].g2; 
	}

	fr_power /= DISP_PIXELS * 4;
	fr_power *= ABL_POW_MULT;
	fr_power >>= ABL_POW_SHIFT;
	calc_cur_power = fr_power;
	
	// Calculate average power using an IIR
	calc_avg_power = ((calc_cur_power * ABL_AVG_PWR_MULT_NEW) + (calc_avg_power * ABL_AVG_PWR_MULT_CUR)) >> ABL_AVG_PWR_SHIFT;
	
	// Calculate OE for ABL limit; peak overrides average, because peak always 
	// triggers at the highest level (lowest OE result)
	if(calc_avg_power > abl_avg_power) {
		abl_limit_avg = (abl_avg_power << DISP_INT_SHIFT) / calc_avg_power;
	} 
	
	if(calc_cur_power > abl_peak_power) {
		abl_limit_peak = (abl_peak_power << DISP_INT_SHIFT) / calc_cur_power;
	}
	
	if(abl_limit_peak < abl_limit_avg) {
		abl_limit_oe = abl_limit_peak;
	} else {
		abl_limit_oe = abl_limit_avg;
	}
	*/
	
	abl_limit_oe = 1023;
	
	// Increase divider count, if the count is reached toggle the LED.
	/*
	disp_led_count++;
	if(disp_led_count == LED_DIVIDE_FSYNC) {
		disp_led_count = 0;
		disp_led_state = 1 - disp_led_state;
		
		if(disp_led_state) 
			GPIOPinWrite(GPIO_PORTN_BASE, GPIO_PIN_0, GPIO_PIN_0);
		else
			GPIOPinWrite(GPIO_PORTN_BASE, GPIO_PIN_0, ~GPIO_PIN_0);
	}
	*/
	
	// Compute the dither data for the frame.  Floyd-Steinberg is applied with the direction
	// of the algorithm changed on each cycle to calculate the correct frame to drive.
	//
	// Dithering increases the perceptible low-level intensity of the panel, improving the rendering
	// at near-black and reducing banding effects.
	disp_dither_calculate();
	
	// Drive the frame.
	disp_drive_frame();
}
	
/*
 * disp_setup: Initialise I/O and SSI modules.
 */
void disp_setup()
{
	uint32_t block, x, y, n;
	
	// ** Setup quad SSI **
	// Enable SSI0 clocks after resetting module
	SysCtlPeripheralDisable(SYSCTL_PERIPH_SSI0);
	SysCtlPeripheralReset(SYSCTL_PERIPH_SSI0);
	SysCtlPeripheralEnable(SYSCTL_PERIPH_SSI0);
	while(!(SysCtlPeripheralReady(SYSCTL_PERIPH_SSI0)));
	
	// Init GPIO
	GPIOPinConfigure(GPIO_PA2_SSI0CLK);
	GPIOPinConfigure(GPIO_PA3_SSI0FSS);
	GPIOPinConfigure(GPIO_PA4_SSI0XDAT0);
	GPIOPinConfigure(GPIO_PA5_SSI0XDAT1);
	GPIOPinConfigure(GPIO_PA6_SSI0XDAT2);
	GPIOPinConfigure(GPIO_PA7_SSI0XDAT3);
  GPIOPinTypeSSI(GPIO_PORTA_BASE, GPIO_PIN_2 | GPIO_PIN_3 | GPIO_PIN_4 | GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7);
	
	// Set pin drive current
	// Clock has highest drive, data lines are lower
	GPIOPadConfigSet(GPIO_PORTA_BASE, GPIO_PIN_2, GPIO_STRENGTH_4MA, GPIO_PIN_TYPE_STD);
	GPIOPadConfigSet(GPIO_PORTA_BASE, GPIO_PIN_3 | GPIO_PIN_4 | GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7, GPIO_STRENGTH_4MA, GPIO_PIN_TYPE_STD);
	//HWREG(GPIO_PORTA_BASE + GPIO_O_DR8R) |= (GPIO_PIN_2 | GPIO_PIN_3 | GPIO_PIN_4 | GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7);
		
	// Init SSI0 module at LED clock rate
	SSIConfigSetExpClk(SSI0_BASE, CPU_CLOCK, SSI_FRF_MOTO_MODE_0, SSI_MODE_MASTER, LED_CLOCK, 8);
	
	// Enable advanced SSI0 mode (quad SSI)
	SSIAdvModeSet(SSI0_BASE, SSI_ADV_MODE_WRITE);
	SSIAdvFrameHoldEnable(SSI0_BASE);
	HWREG(SSI0_BASE + SSI_O_CR1) |= SSI_CR1_HSCLKEN;
	SSIEnable(SSI0_BASE);
	
	// ** Setup GPIO outputs **
	// Global RGB OE.  We go for separate control over R/G/B which gives us fine colour balance control.
	// These are driven by fast interrupts.
	GPIOPinTypeGPIOOutput(DISPCTL_R_OE_PORT, DISPCTL_R_OE_BIT);
	GPIOPinTypeGPIOOutput(DISPCTL_G_OE_PORT, DISPCTL_G_OE_BIT);
	GPIOPinTypeGPIOOutput(DISPCTL_B_OE_PORT, DISPCTL_B_OE_BIT);
	
	// Disable global RGB OE signals so panel is blank (active LOW enable)
	GPIOPinWrite(DISPCTL_R_OE_PORT, DISPCTL_R_OE_BIT, DISPCTL_R_OE_BIT);
	GPIOPinWrite(DISPCTL_G_OE_PORT, DISPCTL_G_OE_BIT, DISPCTL_G_OE_BIT);
	GPIOPinWrite(DISPCTL_B_OE_PORT, DISPCTL_B_OE_BIT, DISPCTL_B_OE_BIT);
	
	// Row select and address
	GPIOPinTypeGPIOOutput(DISPCTL_ODDEVEN_PORT, DISPCTL_ODDEVEN_BIT);
	GPIOPinTypeGPIOOutput(DISPCTL_ADDR_BIT0_PORT, DISPCTL_ADDR_BIT0_MASK);
	GPIOPinTypeGPIOOutput(DISPCTL_ADDR_BIT1_PORT, DISPCTL_ADDR_BIT1_MASK);
	GPIOPinTypeGPIOOutput(DISPCTL_ADDR_BIT2_PORT, DISPCTL_ADDR_BIT2_MASK);
	
	// Setup latch signals as outputs for each superblock
	for(block = 0; block < DISP_NUM_LATCHES; block++) {
		if(disp_latch_lut[block].port_mask != 0) {
			UARTprintf("set latch to output: 0x%08x  mask  0x%08x\r\n", disp_latch_lut[block].port_base, disp_latch_lut[block].port_mask);
			GPIOPinTypeGPIOOutput(disp_latch_lut[block].port_base, disp_latch_lut[block].port_mask);
		}
	}
	
	// Turn off LED2 after init
	GPIOPinWrite(GPIO_PORTN_BASE, GPIO_PIN_0, ~GPIO_PIN_0);	
	
	// Drive LATCH_SEL_0 low;  we do not currently support multi-display
	GPIOPinTypeGPIOOutput(DISPCTL_LATCH_SEL_0_PORT, DISPCTL_LATCH_SEL_0_PIN);
	GPIOPinWrite(DISPCTL_LATCH_SEL_0_PORT, DISPCTL_LATCH_SEL_0_PIN, ~DISPCTL_LATCH_SEL_0_PIN);	
	
	// ** Setup OE timers **
	// Four timers are used, Timer0A, 0B and 1A, for R, G and B respectively,
	// and Timer1B for global subfield timing.
	// Timers are disabled until required.
	SysCtlPeripheralEnable(SYSCTL_PERIPH_TIMER0);
	SysCtlPeripheralEnable(SYSCTL_PERIPH_TIMER1);
	
	// Timer0
	TimerDisable(TIMER0_BASE, TIMER_A);
	TimerDisable(TIMER0_BASE, TIMER_B);
	TimerConfigure(TIMER0_BASE, TIMER_CFG_SPLIT_PAIR | TIMER_CFG_A_ONE_SHOT | TIMER_CFG_B_ONE_SHOT);
	TimerPrescaleSet(TIMER0_BASE, TIMER_A, DISPCTL_OE_PRESCALER);
	TimerPrescaleSet(TIMER0_BASE, TIMER_B, DISPCTL_OE_PRESCALER);
	TimerLoadSet(TIMER0_BASE, TIMER_A, DISPCTL_OE_MAX_PULSE);
	TimerLoadSet(TIMER0_BASE, TIMER_B, DISPCTL_OE_MAX_PULSE);
	TimerMatchSet(TIMER0_BASE, TIMER_A, 0);
	TimerMatchSet(TIMER0_BASE, TIMER_B, 0);
	IntPrioritySet(INT_TIMER0A, 0x20); // High priority
	IntPrioritySet(INT_TIMER0B, 0x40); // 
	TimerIntEnable(TIMER0_BASE, TIMER_TIMA_TIMEOUT);
	TimerIntEnable(TIMER0_BASE, TIMER_TIMB_TIMEOUT);
	IntEnable(INT_TIMER0A);
	IntEnable(INT_TIMER0B);
	
	// Timer1
	TimerDisable(TIMER1_BASE, TIMER_A);
	TimerDisable(TIMER1_BASE, TIMER_B);
	TimerConfigure(TIMER1_BASE, TIMER_CFG_SPLIT_PAIR | TIMER_CFG_A_ONE_SHOT | TIMER_CFG_B_ONE_SHOT);
	TimerPrescaleSet(TIMER1_BASE, TIMER_A, DISPCTL_OE_PRESCALER);
	TimerPrescaleSet(TIMER1_BASE, TIMER_B, DISPCTL_OE_PRESCALER);
	TimerLoadSet(TIMER1_BASE, TIMER_A, DISPCTL_OE_MAX_PULSE);
	TimerLoadSet(TIMER1_BASE, TIMER_B, DISPCTL_OE_MAX_PULSE);
	TimerMatchSet(TIMER1_BASE, TIMER_A, 0);
	TimerMatchSet(TIMER1_BASE, TIMER_B, 0);
	IntPrioritySet(INT_TIMER1A, 0x60);
	IntPrioritySet(INT_TIMER1B, 0x80); // Lower priority for global SF gen to ensure it fires after SFs have been generated
	TimerIntEnable(TIMER1_BASE, TIMER_TIMA_TIMEOUT);
	TimerIntEnable(TIMER1_BASE, TIMER_TIMB_TIMEOUT);
	IntEnable(INT_TIMER1A);
	IntEnable(INT_TIMER1B);
	
	// ** Setup frame timer **
	// This is a general purpose timer (Timer2A) that operates in the 50 - 150Hz range
	// It fires periodically outputting the appropriate subfields
	// Timer is enabled after display starts running
	SysCtlPeripheralEnable(SYSCTL_PERIPH_TIMER2);
	TimerDisable(TIMER2_BASE, TIMER_A);
	TimerConfigure(TIMER2_BASE, TIMER_CFG_SPLIT_PAIR | TIMER_CFG_A_PERIODIC);
	TimerPrescaleSet(TIMER2_BASE, TIMER_A, DISPCTL_FRGEN_PRESCALE);
	IntPrioritySet(INT_TIMER2A, 0xa0);	// Lower priority than all other tasks; avoids immediate OE collision risk
	TimerIntEnable(TIMER2_BASE, TIMER_TIMA_TIMEOUT);
	IntEnable(INT_TIMER2A);
	
	// ** Setup pointers **
	disp_current_ptr = (struct pix_rgbg_t *)&disp_1;
	disp_working_ptr = (struct pix_rgbg_t *)&disp_2;
	
	// Load flash table
	disp_switch_gamma(0);
	
	/*
	while(1) {
		for(n = 0; n < 8; n++) {
			DISPCTL_SB_ADDR_WRITE(disp_addr_lut[n]);
		}
	}
	*/
	
	/*
	while(1) {
		UARTprintf("LATCH ME\r\n");
		
		for(n = 0; n < 8; n++) {
			DISPCTL_LATCH(disp_latch_lut[n]);
		}
	}
	*/
}

/*
 * disp_dither_calculate:  Calculate dither for the current frame and write the output
 * buffer with the dithered result image. (Converts 16 bit image data to 12 bit.)
 */
#define DISP_TIME_DITHER(x, y, pin, pout, div, divmax, col, sn, dither_thresh) \
	oldval = pin[PIX_AT(x, y)].col; \
	newval = pin[PIX_AT(x, y)].col >> div; \
	error = (oldval - (newval << div)); \
	if((error + (ditherrand(x, y, sn) % ((divmax / 2) + 1)) > (divmax / 2))) newval++; \
	pout[PIX_AT((x), (y))].col = newval;
/*
#define DISP_TIME_DITHER(x, y, pin, pout, div, divmax, col, sn, dither_thresh) \
	pout[PIX_AT((x), (y))].col = pin[PIX_AT((x), (y))].col;
*/

inline uint32_t ditherrand(uint32_t x, uint32_t y, uint32_t sub)
{
	//srand(((sub + (x * 4) + (y * DISP_PIX_WIDTH * 4)) * 99) + (disp_frame * 68));
	//return rand();
	return 0;
}

void disp_dither_calculate()
{
	int32_t x, y, newval, oldval, error;
	
	/*
	for(y = 0; y < DISP_PIX_HEIGHT; y++) {
		for(x = 0; x < DISP_PIX_WIDTH; x++) {
			DISP_TIME_DITHER(x, y, disp_current_ptr, disp_out, 4, 16, r, 0, 256);
			DISP_TIME_DITHER(x, y, disp_current_ptr, disp_out, 4, 16, g1, 1, 256);
			DISP_TIME_DITHER(x, y, disp_current_ptr, disp_out, 4, 16, b, 2, 256);
			DISP_TIME_DITHER(x, y, disp_current_ptr, disp_out, 4, 16, g2, 3, 256);
		}
	}
	*/
}

/*
 * disp_output_subfield:  Output a subfield.
 *
 * @param		sf		The subfield to display
 */
void disp_output_subfield(struct pix_rgbg_t *fb, uint32_t bitmask)
{
	static uint32_t frame = 0;
	unsigned int block, module, total_module = 0;
	volatile uint32_t dly;
	struct disp_xyoff_t refxy;
	struct pix_rgbg_t *ptr;
	int i;
	
	// Macro to output a single SF byte
#define OUTPUT_SF_BYTE(i) \
	HWREG(SSI0_BASE + SSI_O_DR) =  \
					((!!(ptr[i*2+1].r & bitmask)) << 0) | ((!!(ptr[i*2+1].g2 & bitmask)) << 1) | \
					((!!(ptr[i*2+1].b & bitmask)) << 2) | ((!!(ptr[i*2+1].g1 & bitmask)) << 3) | \
					((!!(ptr[i*2].r & bitmask)) << 4) | ((!!(ptr[i*2].g2 & bitmask)) << 5) | \
					((!!(ptr[i*2].b & bitmask)) << 6) | ((!!(ptr[i*2].g1 & bitmask)) << 7);
	
	// Process SUPERBLOCKS, each of 8 or 7 modules, which are addressed in halves of the display
	// according to a complex look up table
	for(block = 0; block < DISP_NUM_SUPERBLOCKS; block++) {
		// Compute module address
		refxy = disp_xyoff[block];
		
		// Hold latch HIGH, receiving data
		DISPCTL_LATCH_HIGH(disp_latch_lut[refxy.latch]);
			
		for(module = 0; module < refxy.num_modules; module++, total_module++) {			
			// For each display module, drive two sets of 64 bit data
			// SSI lines are configured in quad SSI module:
			//   bit 0 = R   -  1st pix
			//   bit 1 = G1  ^
			//   bit 2 = B   ^
			//   bit 3 = G2  ^
			//   bit 4 = R   - 2nd pix
			//   bit 5 = G1  ^
			//   bit 6 = B   ^
			//   bit 7 = G2  ^

			// Set the address for this module.  This masks the module to the latch signal
			// that we are sending.
			DISPCTL_SB_ADDR_WRITE(disp_addr_lut[module]);
			
			// Address upper half by setting ODD/EVEN high
			DISPCTL_ODDEVEN_HIGH();
			
			// Calculate pointer and output bytes
			//ptr += 2*SF_STRIDE;
			ptr = fb + (refxy.x + ((refxy.y + module*2) * 2*SF_STRIDE));
			
			for(i = 0; i < 8; i++) {
				OUTPUT_SF_BYTE(i);
			}
			
			// Wait for the FIFO to empty to proceed
			while(!(HWREG(SSI0_BASE + SSI_O_SR) & SSI_SR_TFE)) ;
			
			// Address lower half by setting ODD/EVEN low
			DISPCTL_CHG_DLY1();
			DISPCTL_ODDEVEN_LOW();
			
			// Calculate new pointer and output bytes
			ptr += 2*SF_STRIDE;
			
			for(i = 0; i < 8; i++) {
				OUTPUT_SF_BYTE(i);
			}
			
			/*
			for (i=0; i<8; i++) {
				HWREG(SSI0_BASE + SSI_O_DR) = 
					((!!(ptr[i*2+1].r & bitmask)) << 0) | ((!!(ptr[i*2+1].g2 & bitmask)) << 1) |
					((!!(ptr[i*2+1].b & bitmask)) << 2) | ((!!(ptr[i*2+1].g1 & bitmask)) << 3) |
					((!!(ptr[i*2].r & bitmask)) << 4) | ((!!(ptr[i*2].g2 & bitmask)) << 5) |
					((!!(ptr[i*2].b & bitmask)) << 6) | ((!!(ptr[i*2].g1 & bitmask)) << 7);
			}
			*/

			// Wait for the FIFO to empty to proceed
			while(!(HWREG(SSI0_BASE + SSI_O_SR) & SSI_SR_TFE)) ;
		}
		
		//DISPCTL_CHG_DLY1();
		
		// Return latch LOW
		DISPCTL_LATCH_LOW(disp_latch_lut[refxy.latch]);
	}
}

/*
 * disp_run:  Start displaying stuff.
 */
void disp_run()
{
	uint32_t n;
	
	SSIAdvModeSet(SSI0_BASE, SSI_ADV_MODE_QUAD_WRITE);
	
	UARTprintf("xsz=%6d, ysz=%6d, bytes=%d\r\n", DISP_PIX_WIDTH, DISP_PIX_HEIGHT, DISP_SUBFIELD_SIZE_BYTES);
	
	// Load timer period and enable the timer
	TimerLoadSet(TIMER2_BASE, TIMER_A, DISPCTL_FRGEN_CALC_FREQ(disp_refresh_rate));
	TimerEnable(TIMER2_BASE, TIMER_A);
}

/*
 * disp_stop:  Stop displaying stuff.
 */
void disp_stop()
{
	// Turn off the timer.  The subfields stop shortly afterwards.
	TimerDisable(TIMER2_BASE, TIMER_A);
}

/*
 * disp_output_frame:  Output one frame with desired subfield timing.
 */
void disp_drive_frame()
{
	// Clear framewait flag
	disp_frame_wait = 0;

	// Drive the first subfield - the subfield driver handles it from there.
	cur_subfield = 0;
	disp_drive_subfield();

	// Increase frame counter
	disp_frame++;
}

/*
 * disp_drive_subfield:  Compute and output one subfield.  Triggered by blue OE rising (off)
 * edge, this will output a subfield, or stop the loop if all subfields have been
 * output for the given intensity.
 */
void disp_drive_subfield()
{
	uint32_t x, y, n = 0, n0 = 0, n1 = 0, j = 0, k = 0, field = 1, oe_max = 0, valid = 1;
	uint8_t byte;
	volatile uint32_t dly;
	
	uint32_t base_width, base_scaled, bitmask = 0, v = 0;
	uint32_t oe_offs_r, oe_offs_g, oe_offs_b;
	
	// OE must be off
	DISPCTL_R_OE_OFF();
	DISPCTL_G_OE_OFF();
	DISPCTL_B_OE_OFF();
	
	// Abort if at max subfield, we need to do another frame
	if(cur_subfield == DISP_SUBFIELDS) {
		cur_subfield = 0;
		return;
	}
		
	bitmask = subfield_divider[cur_subfield];
	
	//disp_output_subfield((uint8_t*)&disp_cur_subfield);
	disp_output_subfield(disp_current_ptr, bitmask);
	
	// Determine maximum OE to use: commanded brightness, or brightness (ABL) limit
	if(oe_ints_global < abl_limit_oe) { 
		oe_max = oe_ints_global;
	} else {
		oe_max = abl_limit_oe;
	}
	
	// Compute field widths; these are used by the ISRs to set up the second OE pulse interrupt
	base_width = subfield_widths[cur_subfield];
	base_scaled = (base_width * oe_max) 								>> DISP_INT_SHIFT;
	oe_width_r = (base_scaled * oe_ints_r) 							>> DISP_INT_SHIFT;
	oe_width_g = (base_scaled * oe_ints_g) 							>> DISP_INT_SHIFT;
	oe_width_b = (base_scaled * oe_ints_b) 							>> DISP_INT_SHIFT;
	oe_width_r = (oe_width_r * DISPCTL_OE_SFTDRV_SCALE) >> DISP_INT_SHIFT;
	oe_width_g = (oe_width_g * DISPCTL_OE_SFTDRV_SCALE) >> DISP_INT_SHIFT;
	oe_width_b = (oe_width_b * DISPCTL_OE_SFTDRV_SCALE) >> DISP_INT_SHIFT;
	oe_offs_r = (oe_width_r * DISPCTL_OE_SFTDRV_R_OFFS) >> DISP_INT_SHIFT;
	oe_offs_g = (oe_width_g * DISPCTL_OE_SFTDRV_G_OFFS) >> DISP_INT_SHIFT;
	oe_offs_b = (oe_width_b * DISPCTL_OE_SFTDRV_B_OFFS) >> DISP_INT_SHIFT;
	oe_width_all = base_width + ((base_width * DISPCTL_OE_OFFSET_ALL) >> DISP_INT_SHIFT);
	oe_base = oe_max;
	
	// Load timers for OE initial pulse (Don't do pulses if field is empty)
	oe_gen = valid;
	oe_r_state = OE_STATE_GOLOW;
	oe_g_state = OE_STATE_GOLOW;
	oe_b_state = OE_STATE_GOLOW;
	TimerLoadSet(TIMER0_BASE, TIMER_A, oe_offs_r);
	TimerLoadSet(TIMER0_BASE, TIMER_B, oe_offs_g);
	TimerLoadSet(TIMER1_BASE, TIMER_A, oe_offs_b);
	TimerLoadSet(TIMER1_BASE, TIMER_B, oe_width_all);
	TimerEnable(TIMER0_BASE, TIMER_A);
	TimerEnable(TIMER0_BASE, TIMER_B);
	TimerEnable(TIMER1_BASE, TIMER_A);
	TimerEnable(TIMER1_BASE, TIMER_B);
	
	// Move to next subfield.
	cur_subfield++;
}

/* 
 * disp_write_pixel:  Write to the pixel array at a coordinate.
 */
void disp_write_pixel(uint32_t x, uint32_t y, struct pix_rgbg_t *pix) 
{
	uint32_t addr;
	uint32_t rr, gg1, gg2, bb;
	
	// ignore pixels out of boundaries (<0 not possible)
	if(x >= DISP_PIX_WIDTH)
		return;
	if(y >= DISP_PIX_HEIGHT)
		return;
	
	// clip values
	rr  = pix->r;
	gg1 = pix->g1;
	bb  = pix->b;
	gg2 = pix->g2;
	
	if(rr > 255)
		rr = 255;
	if(gg1 > 255)
		gg1 = 255;
	if(gg2 > 255)
		gg2 = 255;
	if(bb > 255)
		bb = 255;
	
	addr = x + (y * DISP_PIX_WIDTH);
	
	disp_working_ptr[addr].r  = gamma_table_red[pix->r];
	disp_working_ptr[addr].g1 = gamma_table_green[pix->g1];
	disp_working_ptr[addr].b  = gamma_table_blue[pix->b];
	disp_working_ptr[addr].g2 = gamma_table_green[pix->g2];
}

/* 
 * disp_write_pixel_rgb:  Write to the pixel array at a coordinate quickly (direct arg variant)
 */
void disp_write_pixel_rgb(uint32_t x, uint32_t y, uint32_t r, uint32_t g, uint32_t b) 
{
	uint32_t addr;
	
	// ignore pixels out of boundaries (<0 not possible)
	if(x >= DISP_PIX_WIDTH)
		return;
	if(y >= DISP_PIX_HEIGHT)
		return;
	
	// clip values
	if(r > 255)
		r = 255;
	if(g > 255)
		g = 255;
	if(b > 255)
		b = 255;
	
	addr = x + (y * DISP_PIX_WIDTH);
	
	disp_working_ptr[addr].r  = gamma_table_red[r];
	disp_working_ptr[addr].g1 = gamma_table_green[g];
	disp_working_ptr[addr].b  = gamma_table_blue[b];
	disp_working_ptr[addr].g2 = gamma_table_green[g];
	
	/*
	disp_working_ptr[addr].r  = r * 16;
	disp_working_ptr[addr].g1 = g * 16;
	disp_working_ptr[addr].b  = b * 16;
	disp_working_ptr[addr].g2 = g * 16;
	*/
}

/*
 * disp_write_subpixel:  Write a subpixel (gamma corrected) on the display.
 */
void disp_write_subpixel(uint32_t x, uint32_t y, uint8_t value)
{
	uint32_t xx;
	uint32_t yy;
	uint32_t val, xm, ym;
	uint32_t addr;
	
	xx = x / 2;
	yy = y / 2;
	
	// ignore pixels out of boundaries (<0 not possible)
	if(xx >= DISP_PIX_WIDTH)
		return;
	if(yy >= DISP_PIX_HEIGHT)
		return;
	
	// Compute address
	addr = xx + (yy * DISP_PIX_WIDTH);
	
	// Compute subpixel index
	xm = x & 1;
	ym = y & 1;
	val = xm + (ym * 2);
	
	UARTprintf("addr=%d, xx=%d, yy=%d, xm=%d, ym=%d, val=%d\r\n", addr, xx, yy, xm, ym, val);
	
	switch(val) {
		case 0: 
			disp_working_ptr[addr].g1 = gamma_table_green[value];
			break;
		
		case 2: 
			disp_working_ptr[addr].r = gamma_table_red[value];
			break;
		
		case 1: 
			disp_working_ptr[addr].b = gamma_table_blue[value];
			break;
		
		case 3: 
			disp_working_ptr[addr].g2 = gamma_table_green[value];
			break;
	}
}

/* 
 * disp_fill_rgb:  Fill panel with colour.
 */
void disp_fill_rgb(uint32_t r, uint32_t g, uint32_t b)
{
	uint32_t x, y;
	
	for(x = 0; x < DISP_PIX_WIDTH; x++) {
		for(y = 0; y < DISP_PIX_HEIGHT; y++) {
			disp_write_pixel_rgb(x, y, r, g, b);
		}
	}
}

/*
 * disp_test_pattern_boot:  Boot test pattern.
 */
void disp_test_pattern_boot()
{
	uint32_t i;
	
	disp_fill_rgb(255, 0, 0);
	disp_flush_display(1);
	
	for(i = 20; i > 0; i--) 
		disp_wait_for_frame();
	
	disp_fill_rgb(0, 255, 0);
	disp_flush_display(1);
	
	for(i = 20; i > 0; i--) 
		disp_wait_for_frame();
	
	disp_fill_rgb(0, 0, 255);
	disp_flush_display(1);
	
	for(i = 20; i > 0; i--) 
		disp_wait_for_frame();
	
	disp_fill_rgb(255, 255, 255);
	disp_flush_display(1);
	
	for(i = 20; i > 0; i--) 
		disp_wait_for_frame();
		
	disp_fill_rgb(0, 0, 0);
	disp_flush_display(1);
}

/*
 * disp_debug: Dump some useful info
 */
void disp_debug()
{
	UARTprintf("Frame = %7d   Power = %6d mW peak, %6d mW avg.   ABL = %5d   OEW_scale = %5d\r\n", \
		disp_frame, calc_cur_power, calc_avg_power, abl_limit_oe, oe_base);
}

/*
 * disp_wait_for_frame: Wait for frame end. Blocks until frame committed. 
 * Simple way of syncing to display.
 */
void disp_wait_for_frame()
{
	disp_frame_wait = 1;
	
	// Now wait for ISR to clear this flag
	while(disp_frame_wait) ;
}

/*
 * disp_flush_display:  Flushes current display to output on next frame
 * boundary. (Sets flag triggering flush.)
 *
 * @param		block		if set will delay until frame sync actually happens
 */
void disp_flush_display(uint32_t block)
{
	disp_frame_sync = 1;
	
	if(block) 
		disp_wait_for_frame();
}

/*
 * disp_set_brightness:  Set global brightness (0-255)
 *
 * @param		bright		New brightness figure from 0-255, will be gamma corrected to 0-1023
 */
void disp_set_brightness(uint32_t bri)
{
	if(bri > 255)
		bri = 255;
	
	oe_ints_global = def_gamma_lookup_12b[bri] >> 2;
}

/*
 * disp_set_brightness_direct:  Set global brightness without gamma (0-1023)
 */
void disp_set_brightness_direct(uint32_t bri)
{
	oe_ints_global = bri;
}

/*
 * disp_write_gamma:  Write a gamma value in the gamma table.
 */
void disp_write_gamma(uint32_t index, uint32_t col, uint32_t value)
{
	uint16_t *table;
	
	// Select pointer for table
	switch(col) {
		case 0: table = (uint16_t *)&gamma_lookup_red; break;
		case 1: table = (uint16_t *)&gamma_lookup_green; break;
		case 2: table = (uint16_t *)&gamma_lookup_blue; break;
		default: return;
	}
	
	index &= 0xff;
	table[index] = value;
}

/*
 * disp_switch_gamma:  Switch between internal and user defined gamma tables.
 *
 * @param		mode 0: flash gamma, mode 1: user gamma
 */
void disp_switch_gamma(uint32_t mode)
{
	if(mode) {
		// User gamma mode
		gamma_table_red = (uint16_t *)&gamma_lookup_red;
		gamma_table_green = (uint16_t *)&gamma_lookup_green;
		gamma_table_blue = (uint16_t *)&gamma_lookup_blue;
	} else {
		// Flash gamma mode
		gamma_table_red = (uint16_t *)&def_gamma_lookup_12b;
		gamma_table_green = (uint16_t *)&def_gamma_lookup_12b;
		gamma_table_blue = (uint16_t *)&def_gamma_lookup_12b;
	}
}

/*
 * disp_set_oe: Set OE widths (0-1023); used for gamma correction
 * and white balance.
 */
void disp_set_oe(uint32_t r, uint32_t g, uint32_t b)
{
	oe_ints_r = r;
	oe_ints_g = g;
	oe_ints_b = b;
}
